#pragma once
#define _DUMP

#include <stdlib.h>
#include <cstring>
#include <memory>
#include <cstdio>
#include <chrono>

#define MAX_FILES 256
#define SRC_FILENAME "%s.%03i"


extern char _PID[32];
extern int MyId, NumProcesses, _STACKCOUNT;

static inline cl_ulong align(cl_ulong i, int a) { return(((i + a - 1) / a)*a); };

static inline cl_ulong rotl(const cl_ulong x, int k)
{ return ((x << k) | (x >> (64 - k))); }

float diffstats(const float* F, const float* G, size_t count, float* dest = NULL);
float maxupF(const float*  p, size_t len, int stride = 1, float mm = 0.0f, size_t* retIdx = NULL, float* retMean = NULL);

typedef char STR64[64];

inline cl_ulong Microseconds(void)
{
	cl_ulong us = std::chrono::duration_cast<std::chrono::microseconds>
		(std::chrono::high_resolution_clock::now().time_since_epoch()).count();
	return us;
}

struct RectWin
{
	char* ptr;
	cl_uint LWidth, Height, WidthB, BPPixel;

	RectWin() {};

	inline cl_uint WidB(void)  const { return ((LWidth * BPPixel) >> 3); }

	inline char* line(cl_uint Y) const
	{
		Y = __min(Y, Height - 1);
		return ptr + Y * WidthB;
	}

	RectWin(void* p, cl_uint w, cl_uint h, cl_uint bpp, cl_uint lw = 0)
	{
		ptr = (char*)p; LWidth = w; Height = h; BPPixel = bpp;
		WidthB = lw ? lw : WidB();
	}

	void CopyTo(const RectWin& Dest) const
	{
		const cl_uint DW = Dest.WidthB ? Dest.WidthB : WidthB,
			WB = WidB();
		char* pDest = Dest.ptr, *pSrc = ptr;
		for ( cl_uint H = Height; H; --H )
		{
			memmove(pDest, pSrc, WB);
			pSrc += WidthB; pDest += DW;
		}
	}

	void copymem(char* pDest, cl_uint WidthB, bool upload) const
	{
		RectWin DW(pDest, LWidth, Height, BPPixel, WidthB ? WidthB : WidB());
		if ( upload ) CopyTo(DW); else DW.CopyTo(*this);
	}

	size_t copyfile(FILE* fp, bool upload) const
	{
		const cl_uint WB = WidB();
		size_t cnt = 0, ret = 0;
		char* pSrc = ptr;
		if ( !fp ) return 0;

		for ( cl_uint H = Height; H; --H )
		{
			cnt += upload ? fwrite(pSrc, 1, WB, fp) : fread(pSrc, 1, WB, fp);
			ret += cnt;
			if ( cnt < WB ) break;
			pSrc += WidthB;
		}
		return ret;
	}

	template<typename T> float Compare(const RectWin& Dest) const
	{
		const cl_uint DW = Dest.WidthB ? Dest.WidthB : Dest.WidB(),
			XC = DW / sizeof(T);
		char* pSrc = ptr, *pDest = Dest.ptr;
		float ret = 0;

		for ( cl_uint H = Height; H; --H )
		{
#pragma simd
			for ( cl_uint X = XC; X; --X )
			{
				const T a = ((T*)pSrc)[X], b = ((T*)pDest)[X];
				float d = (float)(a - b);	d = (d < 0) ? -d : d;
				ret = __max(ret, d);
			}
			pSrc += WidthB; pDest += DW;
		}
		return ret;
	}

	template<typename T> inline float Compare(const T* Dest) const
	{
		const RectWin D((char*)Dest, LWidth, Height, 32);
		return Compare<T>(D);
	}

}; // struct


template<class T> void QSort(T* a, int lo, int hi)
{
	while ( lo < hi )
	{
		int l = lo, h = hi;  T t, p = a[hi];
		do
		{
			while ( (l < h) && (a[l] <= p) ) l++;
			while ( (h > l) && (p <= a[h]) ) h--;
			if ( l < h ) { t = a[l]; a[l] = a[h]; a[h] = t; }
		} while ( l < h );

		t = a[l]; a[l] = a[hi]; a[hi] = t;
		QSort<T>(a, lo, l - 1);
		lo = l + 1;
	}
}


struct xsrng
{
	cl_ulong S0, S1;
	inline xsrng() { S0 = 0x9876EDBCA9876543; S1 = 0x789ABCDEF123ABCD; }
	inline void seed(cl_ulong s0, cl_ulong s1) { S0 = s0; S1 = s1; (*this)(); }
	inline xsrng(cl_ulong s0, cl_ulong s1) { seed(s0, s1); }

	inline volatile float operator() (void)
	{
		constexpr float scale = 1.0f / (float)(1ULL << 32);
		//int scale_i = 0x2F800000; // float 2^(-32)		
		cl_ulong  s0 = S0, s1 = S1, R = rotl(s0 + s1, 17) + s0;
		s1 ^= s0;
		S0 = rotl(s0, 49) ^ s1 ^ (s1 << 21); // a, b
		S1 = rotl(s1, 28); // c
		return (float)((R >> 13) & 0xFFFFFFFFULL)*scale;
	}

	inline operator float(void) { return (*this)(); }
};

extern const char* ErrName(cl_int n);

#define _MCHK_ 0xAA553C96


#ifdef TARGET_WINDOWS

#include <windows.h>
HANDLE LOPEN(const char* name, int flags);
int LSEEK(HANDLE fd, cl_long offs, int mode);
DWORD LWRITE(HANDLE fd, void* buf, DWORD len);
DWORD LREAD(HANDLE fd, void* buf, DWORD len);
#define LCLOSE CloseHandle
void sleep(int sec);

#else

#ifndef _cdecl
#define _cdecl
#endif

typedef int HANDLE;
constexpr HANDLE INVALID_HANDLE_VALUE = -1;
#define LSEEK ::lseek
#define LWRITE ::write
#define LREAD ::read
#define LOPEN ::open
#define LCLOSE ::close
#define CloseHandle ::close


#endif


extern size_t _cdecl ReturnRIP(size_t s = 0);

cl_uint2 SelectPart(cl_uint N, cl_int pid=-1);

char* FRead(const char* name, size_t* pLen = NULL, bool zero_pad=false);
void FWrite(const char* name, void* r, size_t len, bool append = false);

size_t LOAD(const char* name, void* buf, size_t size, int time = -1);
void DUMP(const char* name, void* buf, size_t size, int time = -1);
void DumpRect(const char* name, void* buf, size_t height, size_t width, size_t stride);

void memswap32(cl_uint* p, cl_long nDwords);

size_t memcmp_ushort_tol(const void* p, const void* q, size_t len, int tol = 1);
int diff_ushort(const void* p, const void* q, size_t len);
size_t memcmp_float_tol(const float* p, const float* q, size_t len, float tol = 1e-5);

char* checked_malloc(size_t s, const char* name = NULL);

bool checked_free(void* data, size_t len = 0, const char* name = NULL);


struct filemap
{
	cl_long len; char* access;
	HANDLE fd; HANDLE mapd;  int error, log;
	inline void init()
	{
		access = NULL; len = 0; log = 0;
		IFLOG(LOG_FILES) { log = 1; }
		error = 0; fd = mapd = INVALID_HANDLE_VALUE;
	}
	inline filemap() { init(); }
	inline operator char*() { return access; }
	inline void Dump(const char* name) { if ( access ) FWrite(name, access, len); }
	virtual void print(const char* prefix = NULL);
	int init(const char* name, cl_long l = -1LL, int myid = 0); // name, len -> file mmap;
	int init(HANDLE fd, int myid = 0); // file-access mode
	void close(void);
	inline filemap(const char* name, cl_long l, int myid = 0) { fd = mapd = INVALID_HANDLE_VALUE; len = 0; access = NULL; error = 0; log = 0;  init(name, l, myid); }
	inline ~filemap() { close(); }
};// struct


struct FileMapArray
{
	filemap Fm[MAX_FILES];
	STR64 ActiveName;
	unsigned int StartNumber;
	FileMapArray();
	virtual int init(unsigned int fst, unsigned int l, cl_long sz, const char* name = NULL, unsigned int startn=0);
	virtual int assert(unsigned int fst, unsigned int l, cl_long sz, size_t MinReq, const char* name, unsigned int startn = 0);
	virtual void close(int i = -1);
	virtual void commit(void);
	virtual int NumOpen(unsigned int fst, unsigned int l, cl_long sz);
	virtual ~FileMapArray() { close(); }
	virtual filemap& operator[](int i) { if ( (i >= 0) && (i < MAX_FILES) ) return Fm[i]; else return Fm[0]; }
};


struct memblk
{
	static constexpr int NAMELEN = 16;
	static size_t ctr;

	char* data; size_t len;
	char name[NAMELEN];

	char* alloc(size_t s, const char* nm = NULL);

	bool  free(void);
	inline char* realloc(size_t s) { if ( s > len ) { if ( data ) free(); return alloc(s); } else return data; }
	//inline operator char*(void) { return data; }
	size_t ReadFile(const char* filename);
	void WriteFile(const char* filename, size_t lim = 0x3FFFFFFFFFFFLL);

	inline bool includes(cl_ulong addr, cl_ulong size) const
	{ return (data && (addr >= (cl_ulong)data) && (addr < (cl_ulong)data + len) && (addr + size < (cl_ulong)data + len)); }

	inline memblk() { data = NULL; len = 0; *name = 0; }
	inline memblk(const char* filename) { data = NULL; len = 0; *name = 0;  ReadFile(filename); }
	inline memblk(size_t s, const char* nm = NULL)
	{
		data = NULL; len = 0; *name = 0;
		if ( nm ) strncpy(name, nm, NAMELEN - 1); else sprintf(name, "MBLK%04X", ctr++);
		alloc(s);
	}
	inline ~memblk() { free(); }
	inline void clr() { if ( data ) memset(data, 0, len); }

	//inline operator bool() { return data };

	inline void Print(void) const
	{
		static constexpr cl_ulong mask = 0x0FFFFFFFFFFFFULL;
		printf("%s%s ->%012llX  end: %012llX", _PID, name, mask&(cl_ulong)data, mask&(cl_ulong)data + len);
	}

};




struct OPTDEF
{
	char name[20]; int type, rep, offs;
	int ParseArgv(int argc, char** argv, const char* base, int argc_pos) const;
	int ParseFile(const char* name, const char* base) const;
	void Print(const char* base, FILE* fp=stdout) const;
	//void Print(const char* base, char* Str) const;
	static constexpr int MAXTYPE = 5, T_INT = 0, T_FLOAT = 1, T_DOUBLE = 2, T_STR64 = 3, T_COMMENT = 4;
	static int   T_SIZE[MAXTYPE];
	static const char*  T_FMT[MAXTYPE];
};

inline bool IsSpace(char c)
{
	return (c == ' ' || c == '\n' || c == '\t' || c == '\r' || c == ',');
}


struct PBMHEADER
{
	static constexpr int MONO16 = 0x15, RGB16 = 0x16, MONO8 = 0x05, RGB8 = 0x06, _16BIT = 0x10;
	char str[24];//		
	PBMHEADER(int W, int H, int color = 0) { Init(W, H, color); }

	void Init(int W, int H, int  color = 0)
	{
		memset(str, 0x0A, sizeof(str));
		if ( (W < 0) || (W > 999999) || (H < 0) || (H > 999999) ) return;
		snprintf(str, sizeof(str), "P%c""\x0A\x0A""%6i""\x0A""%6i""\x0A""%5i",
						 '0' + (color & 0x0F), W, H, (color&_16BIT) ? 65535 : 255);
		str[sizeof(str) - 1] = 0x0A;
	}
};




#ifdef TARGET_WINDOWS

inline HANDLE LOPEN(const char* name, int flags)
{
	return (HANDLE)OpenFile(name, NULL, flags);
}

inline int LSEEK(HANDLE fd, cl_long offs, int mode)
{
	long ll = (offs >> (8 * sizeof(long)));
	return SetFilePointer(fd, offs&(-1LL >> (8 * (sizeof(cl_long) - sizeof(long)))), &ll, mode);
}

inline DWORD LWRITE(HANDLE fd, void* buf, DWORD len)
{
	DWORD res;
	WriteFile(fd, buf, len, &res, NULL);
	return res;
}

inline DWORD LREAD(HANDLE fd, void* buf, DWORD len)
{
	DWORD res;
	ReadFile(fd, buf, len, &res, NULL);
	return res;
}
#endif

cl_long CheckFile(const char* fname);






#define _isInfPtr(p) 0
//( (~(*(cl_uint*)(p))) & 0xEF800000)

void DumpFloatImagePBM(const char* fname, const RectWin& RW, float lim = 1.0f);

static inline void PrintVec(const float4& f, const char* name)
{ printf("\n%s: %f %f %f %f ", name, f[0], f[1], f[2], f[3]); }


int copyname(char* dest, const char* src, int nmax);

struct MATHeader
{
	static const char NameForType[16][8];
	static const int ClassForType[16];
	static const int LenForType[16];
	static constexpr int
		miINT8 = 1, miUINT8 = 2, miINT16 = 3, miUINT16 = 4, miINT32 = 5, miUINT32 = 6,
		miINT64 = 12, miUINT64 = 13, miMATRIX = 14, miSINGLE = 7, miDOUBLE = 9,
		mxSINGLE_CLASS = 7, mxDOUBLE_CLASS = 6, mxINT8_CLASS = 8, mxUINT8_CLASS = 9,
		mxINT16_CLASS = 10, mxUINT16_CLASS = 11, mxINT32_CLASS = 12, mxUINT32_CLASS = 13,
		mxINT64_CLASS = 14, mxUINT64_CLASS = 15;


	char desc[116];
	cl_int sub1 = 0, sub2 = 0; // unaligned long
	cl_short version = 0x100, endian = 'MI';

	cl_uint MatrixHead = miMATRIX, MatrixLen,
		aflagsHead = miUINT32, aflagsLen = 8; int aflags1 = 0, aflags2 = 0;
	cl_uint DimHead = miINT32, DimLen = 16, Dim[4] = { 1,1,1,1 },
		NameHead = miINT8, NameLen = 64;
	STR64 Name; cl_ulong _pad;

	cl_uint DataType, DataLen,
		ActNameLen, HeadSize;

	MATHeader() {};

	void Create(const int* pDim, int Type, const char* name);
	MATHeader(const int* pDim, int Type, const char* name)
	{ Create(pDim, Type, name); };

	FILE* WriteFile(void* data = NULL, const char* fname = NULL) const;

	FILE* WriteFile(const int* pDim, int Type, const char* name, void* data = NULL, const char* fname = NULL)
	{ Create(pDim, Type, name); return WriteFile(data, fname); }

};

