
typedef __global char *GPTR;
typedef __local  char *LocPTR;
typedef __constant CONST_DATA *CPTR;


#define GI_(a) (*(__global int*)(a))
#define GU_(a) (*(__global uint*)(a))
#define GF_(a) (*(__global float*)(a))
#define GL_(a) (*(__global long*)(a))

#define GFp_  (__global float*)
#define GSp_  (__global short*)


__kernel void CheckData(CPTR CD, ulong ReturnPtr)
{
	__global ulong *p = (__global ulong*)ReturnPtr;
	p[0] = CD->pgLU.s0;
	p[1] = CD->pgLU.s1;
	p[2] = CD->pgLW.s0;
	p[3] = CD->pgLW.s1;
	p[4] = CD->pgGrad;
	p[5] = CD->pgCoeff;
	p[6] = CD->DZ;
}

__kernel void GetAddr(GPTR Data)
{	
	GL_(Data) = (ulong)Data;
}

__kernel void clw(GPTR Data, uint offs, uint n)
{
	ulong pos = ((get_global_id(0) + get_global_id(1)*get_global_size(0) + get_global_id(2)*get_global_size(0)*get_global_size(0)) << 2)*n;
	pos = (ulong)Data + pos + offs;
	for ( int i = n - 1; i >= 0; --i ) { GF_(pos) = 0; pos += 4; }
}// end kernel

__kernel void CLS(ulong Data)
{
	ulong Base = Data + ((get_global_id(0) + get_global_size(0)*get_global_id(1)) << 4);
	GL_(Base) = 0L; GL_(Base + 8) = 0L;
}

__kernel void CLS2(ulong Data, long N)
{
	ulong Base = Data + (get_global_id(0) << 2);
	ulong DY = (get_global_size(0) << 2);
	long i;
	for ( i = N; i > 0; --i )
	{ GI_(Base) = 0;  Base += DY; }
}



__kernel void NRBCond(CPTR CD, ulong pos, int Xdir, int Ydir, int Wincr)
{
	//return;
	pos += (long)((int)get_global_id(0)*Xdir);
	int YY = Ydir + Ydir, YYY = YY + Ydir;
	ulong pDwnLo = pos + YY + YY;
	bool zz = (get_global_id(0) < 4 || get_global_id(0) >= get_global_size(0) - 4);
	float MIDHI, MIDLO,
		DIFF_ = GF_(pDwnLo - Xdir) + GF_(pDwnLo + Xdir) +
		GF_(pos - Xdir) + GF_(pos + Xdir);

	float out = zz ? 0.0f : CD->Kuh*GF_(pos) + // UP HI
		CD->Kdh*GF_(pos + YY) + //DWN HI
		CD->Kdl*GF_(pDwnLo) + // DWN LO 
		CD->Kmid*((MIDHI = GF_(pos + Wincr)) + (MIDLO = GF_(pos + Wincr + Ydir))) + // MIDHL_=s(imid,HI,:)+s(imid,LO,:);
		CD->Kdif * DIFF_;
	//out = get_global_id(0); // 0.0f;
	GF_(pos + Wincr + YY) = MIDHI;
	GF_(pDwnLo + Wincr) = MIDLO;
	GF_(pos + Ydir) = out;
	GF_(pos + YY) = out * 0.75;
	GF_(pos + YYY) = 0.0f;
}

#define NRBC_ZEROING_WIDTH 1

__kernel void NRBCond_Hrz(CPTR CD, ulong pos, int Wincr)
{ // NDR X=width Y=2 Z=NSrc
	//return;
	int Ydir = -CD->DY;
	if ( get_global_id(1) ) 	Ydir = -Ydir;

	pos += (long)((int)get_global_id(0) * 4) + get_global_id(2)*CD->DZ
		+ get_global_id(1)*CD->DY * (CD->InH - 1);

	int YY = Ydir + Ydir, YYY = YY + Ydir;
	ulong pDwnLo = pos + YY + YY;
	bool zz = (get_global_id(0) < NRBC_ZEROING_WIDTH || get_global_id(0) >= get_global_size(0) - NRBC_ZEROING_WIDTH);
	float MIDHI, MIDLO,
		DIFF_ = GF_(pDwnLo - 4) + GF_(pDwnLo + 4) +
		GF_(pos - 4) + GF_(pos + 4);

	float out = CD->Kuh*GF_(pos) + // UP HI
		CD->Kdh*GF_(pos + YY) + //DWN HI
		CD->Kdl*GF_(pDwnLo) + // DWN LO 
		CD->Kmid*((MIDHI = GF_(pos + Wincr)) + (MIDLO = GF_(pos + Wincr + Ydir))) + // MIDHL_=s(imid,HI,:)+s(imid,LO,:);
		CD->Kdif * DIFF_;
	//out = get_global_id(0); // 0.0f;

	if ( zz ) out = 0;
	{
		GF_(pos + Wincr + YY) = MIDHI;
		GF_(pDwnLo + Wincr) = MIDLO;
		GF_(pos + Ydir) = out;
		GF_(pos + YY) = out * 0.75;
		GF_(pos + YYY) = 0.0f;
	}
}

__kernel void NRBCond_Vrt(CPTR CD, ulong pos, int Wincr)
{ // NDR X=width Y=NSrc*2
	//return;
	int Ydir = -4, DY = CD->DY;
	if ( get_global_id(1) ) 	Ydir = -Ydir;

	pos += (long)((int)get_global_id(0) * DY) + (get_global_id(2))*CD->DZ
		+ (get_global_id(1)) * 4 * (CD->InW - 1);

	int YY = Ydir + Ydir, YYY = YY + Ydir;
	ulong pDwnLo = pos + YY + YY;
	bool zz = ((get_global_id(0) < NRBC_ZEROING_WIDTH) || (get_global_id(0) >= get_global_size(0) - NRBC_ZEROING_WIDTH));
	float MIDHI, MIDLO,
		DIFF_ = GF_(pDwnLo - DY) + GF_(pDwnLo + DY) +
		GF_(pos - DY) + GF_(pos + DY);

	float out = CD->Kuh*GF_(pos) + // UP HI
		CD->Kdh*GF_(pos + YY) + //DWN HI
		CD->Kdl*GF_(pDwnLo) + // DWN LO 
		CD->Kmid*((MIDHI = GF_(pos + Wincr)) + (MIDLO = GF_(pos + Wincr + Ydir))) + // MIDHL_=s(imid,HI,:)+s(imid,LO,:);
		CD->Kdif * DIFF_;
	if ( zz ) out = 0;

	{

		GF_(pos + Wincr + YY) = MIDHI;
		GF_(pDwnLo + Wincr) = MIDLO;
		GF_(pos + Ydir) = out;
		GF_(pos + YY) = out * 0.75;
		GF_(pos + YYY) = 0.0f;
	}
}

#undef NRBC_ZEROING_WIDTH



__kernel void CompareData(CPTR CD, ulong ExData, ulong ExData2, uint nFrames)
{ // 1-D NDR NDets
	uint SectionWidthB = (get_global_size(0) << 2);

	ulong offs = (get_global_id(0) << 2),
		sumptr = ExData + offs;

	double sum = 0;

	for ( int Z = 0; Z < nFrames; ++Z )
	{		
		double a = (double)GF_(ExData + offs),
			b = (double)GF_(ExData2 + offs);
		sum += (a - b)*(a - b);
		offs += SectionWidthB;
	}
	GF_(sumptr) = convert_float(sum);
}



__kernel void SaveExData(CPTR CD, ulong destptr, int Frame)
{
	if ( (Frame < 1) || (Frame > CD->NFrames) ) return;
#define LOADBOUNDCORE 	uint SectionWidthB = (get_global_size(0) << 2);\
	destptr += (get_global_id(0) << 2);\
	ulong offs = GU_(destptr-SectionWidthB)+CD->DZ*get_global_id(1);\
	destptr += ((Frame-1) + get_global_id(1)*CD->NFrames) * SectionWidthB;\
	ulong srcptr = ((Frame & 1) ? CD->pgLU.y : CD->pgLU.x) + offs;
//-end-macro
	LOADBOUNDCORE
		GF_(destptr) = GF_(srcptr);
}// end kernel

__kernel void LoadBound(CPTR CD, ulong destptr, int Frame)
{
	LOADBOUNDCORE
		GF_(srcptr) = GF_(destptr);
}// end kernel

__kernel void LoadExData(CPTR CD, ulong destptr, int Frame, float nvscale)
{
	int TBase = clamp((int)(Frame -1), (int)1, (int)(CD->NFrames - 1));
	ulong SectionWidthB = (get_global_size(0) << 2);
	destptr += (get_global_id(0) << 2);
	ulong offs = GU_(destptr - SectionWidthB) + CD->DZ*get_global_id(1);
	ulong NevPtr = destptr - SectionWidthB * ((get_global_id(1) + 2));
	destptr += (TBase + get_global_id(1)*CD->NFrames) * SectionWidthB;
	ulong srcptr = ((Frame & 1) ? CD->pgLU.y : CD->pgLU.x) + offs;

	ulong srcW = ((Frame & 1) ? CD->pgLW.y : CD->pgLW.x) +
		(ulong)offs;	
	float nv = (GF_(srcptr) - GF_(destptr))*nvscale;
	GF_(NevPtr) += nv * nv;

	if ( get_global_id(0) < CD->NumDetectors )
		GF_(srcW) += nv;
}// end kernel


#define LAPL(n) (CD->LAPL.s##n)

#define LOADLINE(E)  U2##E=vload4(0,LCenter);        /* V2 = a  b  A  B  */ \
  CC=vload4(0,LCenter+4);                              /* CC = C  D  E  F  */ \
	U0##E.lo=U2##E.hi; U0##E.hi=CC.lo;                  /* V0 = A  B  C  D */ \
	U1##E.s012=U2##E.s123+U0##E.s123; \
	LCenter+=DYi; \
  U1##E.s3=CC.s0+CC.s2;                               /* V1 = bB AC BD CE */ \
	U2##E+=CC; 
//-end-macro


#define PROCESS(_A,_B,_C,_D,_E)	{ LOADLINE(_E); \
	U2##_A += U2##_E; \
   U0##_A =  (U0##_A + U0##_E + U2##_C)*LAPL(3) + U2##_A*LAPL(5); \
		 CAi=vload4(0,pCoeff+imgcoord.y*DYi); \
	 U0##_A += (U1##_C + U0##_B + U0##_D)*LAPL(1);\
 CC=convert_float4(CAi<<16)*CD->CScale;  \
/* CC=(1.0f/1.5f/CD->dt); */ \
   U0##_A +=	(U1##_A + U1##_E + U2##_B + U2##_D)*LAPL(4);\
     AA=convert_float4(CAi>>16)*CD->AScale; \
	 U0##_A +=(U1##_B + U1##_D)*LAPL(2); \
			U2##_A=vload4(0,LDest);   CC*=CC; \
	 ++imgcoord.y;  \
		  U1##_A = (U0##_C * (CC*2.0f+LAPL(0)) + U0##_A - U2##_A* (CC-AA)) / (CC+AA); /* V1=UWnew, V0 free */\
 vstore4(U1##_A,0,LDest);    LDest+=DYi;} 
//-end-macro

#define PROCESSLOOP		LOADLINE(A); LOADLINE(B);		LOADLINE(C); LOADLINE(D);	 \
for ( i = YY; i > 0; i -= 5 ) \
{	PROCESS(A, B, C, D, E); 	PROCESS(B, C, D, E, A); \
  PROCESS(C, D, E, A, B);		PROCESS(D, E, A, B, C); \
  PROCESS(E, A, B, C, D); 		}
//-end-macro

#define SETIMGCOORD int2 imgcoord = (int2)(get_global_id(0) << 2, get_global_id(1)*YY); 
//-end--macro

__kernel void ForwardWave44(CPTR CD, int Frame, int YY)
{
	SETIMGCOORD
		__global uint* pCoeff = ((__global uint*)CD->pgCoeff) + imgcoord.x;

#define _FORWARD_WAVE 	bool oddframe = (Frame & 1);\
	float4 CC, AA, U0A, U1A, U2A, U0B, U1B, U2B, \
		U0C, U1C, U2C, U0D, U1D, U2D, U0E, U1E, U2E; \
	int i;  uint4 CAi; \
	__global float *LCenter, *LDest; \
	uint DYi = (CD->DY>>2); /* floats per vector. ptrs are floats! */ \
	uint Base =  imgcoord.x + imgcoord.y * DYi+get_global_id(2)*(CD->DZ>>2); \
	LDest = GFp_(oddframe? CD->pgLU.x: CD->pgLU.y) + Base;  \  
	LCenter = GFp_(oddframe ? CD->pgLU.y: CD->pgLU.x) + Base - DYi - DYi - 2; \
  PROCESSLOOP	
//-end-macro

	_FORWARD_WAVE

} // end kernel

#undef _FORWARD_WAVE		
#undef PROCESS
#undef LOADLINE




#define LOADLINE(E)  U2##E=vload4(0,LCenter);   W2##E=vload4(0,LWCenter);   /* V2 = a  b  A  B  */ \
  CC=vload4(0,LCenter+4); AA = vload4(0, LWCenter + 4);                        /* CC = C  D  E  F  */ \
	U0##E.lo=U2##E.hi; W0##E.lo = W2##E.hi; U0##E.hi=CC.lo; W0##E.hi = AA.lo;  /* V0 = A  B  C  D */ \
	U1##E.s012=U2##E.s123+U0##E.s123; W1##E.s012=W2##E.s123+W0##E.s123; \
	LCenter+=DYi; LWCenter+=DYi; \
  U1##E.s3=CC.s0+CC.s2; W1##E.s3=AA.s0+AA.s2;                                 /* V1 = bB AC BD CE */ \
	U2##E+=CC; W2##E+=AA; 
//-end-macro



#define COMPUTEGRAD { short8 GS;\
  GS=vload8(0, (__global short*)(pGrad + imgcoord.y*DYi)); \
	GS.s0246= convert_short4_sat_rte( (convert_float4(GS.s0246)+CC*GradMix*CD->GradScale.x) ); /* RTZ - smoother, RTE - precise*/ \
  GS.s1357= convert_short4_sat_rte( (convert_float4(GS.s1357)+AA*GradMix*CD->GradScale.y )); \
	vstore8(GS, 0, (__global short*)(pGrad + imgcoord.y*DYi)); }
	//-end-macro

#define PROCESS(_A,_B,_C,_D,_E)	{  LOADLINE(_E); \
	CAi=vload4(0,pCoeff+imgcoord.y*DYi); \
	 U2##_A += U2##_E; W2##_A += W2##_E; \
   U0##_A =  (U0##_A + U0##_E + U2##_C)*LAPL(3) + U2##_A*LAPL(5); \
	 W0##_A =  (W0##_A + W0##_E + W2##_C)*LAPL(3) + W2##_A*LAPL(5); \
	 U0##_A += (U1##_C + U0##_B + U0##_D)*LAPL(1);\
	 W0##_A += (W1##_C + W0##_B + W0##_D)*LAPL(1);\
  CC=convert_float4(CAi<<16)*CD->CScale;  \
   U0##_A +=	(U1##_A + U1##_E + U2##_B + U2##_D)*LAPL(4);\
   W0##_A +=	(W1##_A + W1##_E + W2##_B + W2##_D)*LAPL(4);\
  AA=convert_float4(CAi>>16)*CD->AScale; \
	 U0##_A +=(U1##_B + U1##_D)*LAPL(2); \
	 W0##_A +=(W1##_B + W1##_D)*LAPL(2); \
	CC*=CC; \
			U2##_A=vload4(0,LDest); W2##_A=vload4(0,LWDest);  \
		  U1##_A = (U0##_C * (CC*2.0f+LAPL(0)) + U0##_A - U2##_A* (CC+AA)) / (CC-AA); /* V1=UWnew, V0 free */\
		  W1##_A = (W0##_C * (CC*2.0f+LAPL(0)) + W0##_A - W2##_A* (CC-AA)) / (CC+AA); /* V1=UWnew, V0 free */\
				vstore4(U1##_A,0,LDest);  vstore4(W1##_A,0,LWDest);\
     CC = (U0##_C - U1##_A) * (W0##_C - W1##_A); \
     AA =  U0##_C           * (W2##_A - W1##_A); \
      COMPUTEGRAD; ++imgcoord.y;  \
			    LDest+=DYi;    LWDest+=DYi;    } 
//-end-macro

__kernel void BackwardWave44s(CPTR CD, int Frame, int YY)
{
	SETIMGCOORD
		__global int* pCoeff = ((__global int*)CD->pgCoeff) + imgcoord.x;
	__global int* pGrad = ((__global int*)CD->pgGrad) + imgcoord.x;

	bool oddframe = (Frame & 1);
	float GradMix = clamp((convert_float(Frame) - CD->EndInvTime) / CD->GradMixDuration, 0.0f, 1.0f);
	float4 CC, AA, U0A, U1A, U2A, U0B, U1B, U2B, U0C, U1C, U2C, U0D, U1D, U2D, U0E, U1E, U2E;
	float4  W0A, W1A, W2A, W0B, W1B, W2B, W0C, W1C, W2C, W0D, W1D, W2D, W0E, W1E, W2E;
	int i;  int4 CAi;
	__global float *LCenter, *LDest, *LWCenter, *LWDest;

	uint DYi = (CD->DY >> 2); /* floats per vector. ptrs are floats! */

	uint Base = (imgcoord.x + imgcoord.y * DYi);
	LDest = GFp_(oddframe ? CD->pgLU.x : CD->pgLU.y) + Base;
	LCenter = GFp_(oddframe ? CD->pgLU.y : CD->pgLU.x) + Base - DYi - DYi - 2;
	LWDest = GFp_(oddframe ? CD->pgLW.x : CD->pgLW.y) + Base;
	LWCenter = GFp_(oddframe ? CD->pgLW.y : CD->pgLW.x) + Base - DYi - DYi - 2;

	LOADLINE(A); LOADLINE(B);		LOADLINE(C); LOADLINE(D);
	for ( i = YY; i > 0; i -= 5 )
	{
		PROCESS(A, B, C, D, E); 	PROCESS(B, C, D, E, A);
		PROCESS(C, D, E, A, B);		PROCESS(D, E, A, B, C);
		PROCESS(E, A, B, C, D);
	}
} // end kernel

#undef COMPUTEGRAD
//LOCAL GRAD
#define COMPUTEGRAD CC*=GradMix; CC += vload4(LocIdx, LOC); \
   AA*=GradMix; AA+=vload4(LocIdx+1,LOC); \
   vstore4(CC,LocIdx,LOC); vstore4(AA,LocIdx+1,LOC); LocIdx+=(get_local_size(0) << 1); 
//-end-macro


__kernel void BackwardWave44(CPTR CD, int Frame, int YY, int ActiveSrcs)
{
	SETIMGCOORD
		__global int* pCoeff = ((__global int*)CD->pgCoeff) + imgcoord.x;
	__global int* pGrad = ((__global int*)CD->pgGrad) + imgcoord.x;

	bool oddframe = (Frame & 1);
	float GradMix = clamp((convert_float(Frame) - CD->EndInvTime) / CD->GradMixDuration, 0.0f, 1.0f);
	float4 CC, AA, U0A, U1A, U2A, U0B, U1B, U2B, U0C, U1C, U2C, U0D, U1D, U2D, U0E, U1E, U2E;
	float4  W0A, W1A, W2A, W0B, W1B, W2B, W0C, W1C, W2C, W0D, W1D, W2D, W0E, W1E, W2E;
	int i;  int4 CAi;
	__global float *LCenter, *LDest, *LWCenter, *LWDest;

	uint DYi = (CD->DY >> 2); /* floats per vector. ptrs are floats! */

	__local float LOC[GPU_LOCAL_MEM_SIZE >> 2];
	uint LocIdx = (get_local_id(0) << 1);
	for ( i = 0; i < YY; ++i, LocIdx += (get_local_size(0) << 1) )
	{ // download gradient array to local memory
		short8 GS = vload8(0, (__global short*)(pGrad + (imgcoord.y + i)*DYi));
		CC = 0; convert_float4(GS.s0246) / CD->GradScale.x;
		AA = 0; convert_float4(GS.s1357) / CD->GradScale.y;
		vstore4(CC, LocIdx, LOC);   vstore4(AA, LocIdx + 1, LOC);
	}


	for ( int CurSou = 0; CurSou < ActiveSrcs; ++CurSou )
	{
		uint Base = (imgcoord.x + imgcoord.y * DYi) + CurSou * (CD->DZ >> 2);;
		LDest = GFp_(oddframe ? CD->pgLU.x : CD->pgLU.y) + Base;
		LCenter = GFp_(oddframe ? CD->pgLU.y : CD->pgLU.x) + Base - DYi - DYi - 2;
		LWDest = GFp_(oddframe ? CD->pgLW.x : CD->pgLW.y) + Base;
		LWCenter = GFp_(oddframe ? CD->pgLW.y : CD->pgLW.x) + Base - DYi - DYi - 2;

		LocIdx = (get_local_id(0) << 1);
		LOADLINE(A); LOADLINE(B);		LOADLINE(C); LOADLINE(D);
		for ( i = YY; i > 0; i -= 5 )
		{
			PROCESS(A, B, C, D, E); 	PROCESS(B, C, D, E, A);
			PROCESS(C, D, E, A, B);		PROCESS(D, E, A, B, C);
			PROCESS(E, A, B, C, D);
		}
		imgcoord.y -= YY;
	}// for CurSou

	LocIdx = (get_local_id(0) << 1);
	for ( i = 0; i < YY; ++i, LocIdx += (get_local_size(0) << 1) )
	{ // offload local memory to gradient array
		CC = vload4(LocIdx, LOC);   AA = vload4(LocIdx + 1, LOC);		 		
		uint4 GS = (convert_uint4(as_ushort4(convert_short4_sat_rte((CC*CD->GradScale.x)))))
		       	|  (convert_uint4(as_ushort4(convert_short4_sat_rte((AA*CD->GradScale.y)))) << 16);
		vstore4(GS, 0, (__global uint*)(pGrad + (imgcoord.y + i)*DYi));		
	}
} // end kernel


__kernel void SumParGrad(CPTR CD, int ActiveSrcs)
{
#define YY (1)
	SETIMGCOORD
#undef YY
	ulong pOutGrad =  get_global_id(1)*CD->DY + (get_global_id(0)<<4);
	ulong pPGrad = CD->pgParGrad + pOutGrad, DZ= CD->DY*CD->FullH;
	pOutGrad += CD->pgGrad;
	
	float8 FG = 0.0f; short8 SG;
	for ( int i = 0; i < ActiveSrcs; ++i )
	{
		SG = vload8(0, GSp_(pPGrad));	
		FG += convert_float8(SG);
		SG = 0;
		vstore8(SG, 0, GSp_(pPGrad));
		pPGrad += DZ;
	}
	SG=vload8( 0, GSp_(pOutGrad));
	FG.s0246 *= CD->GradScale.x / CD->ParGradScale.x;
	FG.s1357 *= CD->GradScale.y / CD->ParGradScale.y;
	FG += convert_float8(SG); 	
	SG = convert_short8_sat_rte(FG);
	vstore8(SG, 0,GSp_(pOutGrad));
}

#undef COMPUTEGRAD

#define COMPUTEGRAD { float8 GF;\
  GF=convert_float8(vload8(0, (__global short*)(pGrad + imgcoord.y*DYi))); \
	GF.s0246+= (CC*GradMix*CD->ParGradScale.x); /* RTZ - smoother, RTE - precise*/ \
  GF.s1357+= (AA*GradMix*CD->ParGradScale.y); \
	vstore8(convert_short8_sat_rte(GF), 0, (__global short*)(pGrad + imgcoord.y*DYi)); }
	//-end-macro


__kernel void BackwardWave44pg(CPTR CD, int Frame, int YY)
{
	SETIMGCOORD
		__global int* pCoeff = ((__global int*)CD->pgCoeff) + imgcoord.x;
	__global int* pGrad = ((__global int*)(CD->pgParGrad + get_global_id(2) *CD->DY*CD->FullH)) + imgcoord.x;

	bool oddframe = (Frame & 1);
	float GradMix = clamp((convert_float(Frame) - CD->EndInvTime) / CD->GradMixDuration, 0.0f, 1.0f);
	float4 CC, AA, U0A, U1A, U2A, U0B, U1B, U2B, U0C, U1C, U2C, U0D, U1D, U2D, U0E, U1E, U2E;
	float4  W0A, W1A, W2A, W0B, W1B, W2B, W0C, W1C, W2C, W0D, W1D, W2D, W0E, W1E, W2E;
	int i;  int4 CAi;
	__global float *LCenter, *LDest, *LWCenter, *LWDest;

	uint DYi = (CD->DY >> 2); /* floats per vector. ptrs are floats! */

	uint Base = (imgcoord.x + imgcoord.y * DYi) + get_global_id(2) * (CD->DZ >> 2);;
	LDest = GFp_(oddframe ? CD->pgLU.x : CD->pgLU.y) + Base;
	LCenter = GFp_(oddframe ? CD->pgLU.y : CD->pgLU.x) + Base - DYi - DYi - 2;
	LWDest = GFp_(oddframe ? CD->pgLW.x : CD->pgLW.y) + Base;
	LWCenter = GFp_(oddframe ? CD->pgLW.y : CD->pgLW.x) + Base - DYi - DYi - 2;
	
	LOADLINE(A); LOADLINE(B);		LOADLINE(C); LOADLINE(D);
	for ( i = YY; i > 0; i -= 5 )
	{
		PROCESS(A, B, C, D, E); 	PROCESS(B, C, D, E, A);
		PROCESS(C, D, E, A, B);		PROCESS(D, E, A, B, C);
		PROCESS(E, A, B, C, D);
	}
} // end kernel


#undef LAPL
#undef PROCESS
#undef LOADLINE


#define APPROX_ARCCOS(fade) (((( 0.074261-0.018793*fade)*fade - 0.2121144)*fade + 1.5707288)*sqrt(fmax(0.0f,1.0f - fade)))

__kernel void Initialize(CPTR CD,int StartSrc)
{ // NDR: 3D InW*InH*NSources
	float2 CXY = (float2)(get_global_size(0)*0.5f, get_global_size(1)*0.5f),
		SrcXY = CD->SrcPos[get_global_id(2)+StartSrc],
		Vec = (float2)(get_global_id(0), get_global_id(1)) - SrcXY,
		Orth = CXY - SrcXY; 	Orth /= length(Orth);

	float WaveShift = CD->dt / CD->dx * CD->Vel;

	float BegImp = CD->ImpStart / CD->dx,
		WidthImp = CD->ImpWidth / CD->dx,
		EndImp = BegImp + WidthImp,
		Amp = 100.0f / sqrt(CD->ImpWidth), // sqrt(0.5f*(EndImp + BegImp))*3.0f,
		RWidthImp = 1.0f / WidthImp;

	float R = fmax(length(Vec), 1.0f), fade0 = dot(Vec, Orth) / R,
	fade = (M_PI - (fade0=APPROX_ARCCOS(fade0))) / M_PI;
	fade = clamp(fade, CD->ImpClipAMin, CD->ImpClipAMax) - CD->ImpClipAMin;
	fade /= (CD->ImpClipAMax - CD->ImpClipAMin);
	//fade = 1.0f;
	fade *= Amp / sqrt(R);

	float f0 = R - BegImp, f1 = f0 - WaveShift;
	f0 = clamp(f0, 0.0f, WidthImp)*RWidthImp;
	f1 = clamp(f1, 0.0f, WidthImp)*RWidthImp;
	f0 = sinpi(f0)*sinpi(f0*CD->ImpFreq)*fade;
	f1 = sinpi(f1)*sinpi(f1*CD->ImpFreq)*fade;

	ulong Base = (get_global_id(0) << 2) + get_global_id(1)*CD->DY + get_global_id(2)*CD->DZ;

	*(__global float*)(CD->pgLU.x + Base) = f0;
	*(__global float*)(CD->pgLU.y + Base) = f1;


}

