#pragma once

#include "setup.h"
#define _USE_MATH_DEFINES

#define DEBUG

#define BREAKPOINT {*(int*)(0) = 0;};
#ifdef DEBUG
#define CHECKPOINT(str) { IFLOG(LOG_CHECKS) { printf("%s%s ",_PID,(str)); fflush(stdout); } }
#define PAUSE(str) { puts(str); fflush(stdout); getchar(); }
#else
#define CHECKPOINT(str)
#define PAUSE(str)
#endif


#ifdef TARGET_GPU
#include "CL/cl.h"
#else
#include <stdint.h>
typedef int16_t  cl_short;
typedef uint16_t cl_ushort;
typedef int32_t  cl_int;
typedef uint32_t cl_uint;
typedef int64_t  cl_long;
typedef uint64_t cl_ulong;
typedef uint64_t cl_mem;
typedef struct { float s[2]; } cl_float2;
typedef struct { int s[4]; }   cl_int4;
typedef struct { unsigned int s[2]; }   cl_uint2;
typedef struct { short s[2]; } cl_short2;
typedef struct { unsigned short s[2]; } cl_ushort2;
typedef struct { cl_ulong s[2]; } cl_ulong2;
typedef struct { unsigned short s[8]; } cl_ushort8;
typedef struct { float s[8]; } cl_float8;

typedef int cl_command_queue;
//typedef struct { float s[4]; } cl_float4;
//typedef struct { cl_ulong s[4]; } cl_ulong4;
//typedef struct { cl_uint s[4]; } cl_uint4;
//typedef struct { float s[8]; } cl_float8;
//typedef float cl_float;
#endif

#ifdef __GNUC__
#define _cdecl __attribute__((__cdecl__))
#endif // __GNUC__

#ifdef TARGET_ARM
#include "arm_vcl.h"
#else
#include "vectorclass/vectorclass.h"
typedef Vec4i int4;  typedef Vec4ui uint4;   typedef Vec4f float4;
typedef Vec8ui uint8; typedef Vec8i int8;   typedef Vec8f float8;
typedef Vec16i int16; typedef Vec16ui uint16;
typedef Vec16f float16; typedef Vec16s short16;
typedef Vec32s short32; typedef Vec8s short8;
#endif

#ifndef __min
#define __min(a,b) ((a)<(b)? (a):(b))
#endif

#ifndef __max
#define __max(a,b) ((a)>(b)? (a):(b))
#endif

#define CLAMP(v,lo,hi) ( ((v)<(lo))? (lo) : ( ((v)>(hi))? (hi) : (v)) )

enum
{
	LOG_INFO = 1, LOG_FMAP = 2, LOG_MALLOC = 4, LOG_CHECKS = 8,
	LOG_GPU = 0x10, LOG_CHECKS2 = 0x20, LOG_MULTI = 0x40, LOG_TSTEP = 0x80,
	LOG_DUMP = 0x100, LOG_STATS = 0x200, LOG_KRNL = 0x400, LOG_FILES = 0x800,
	LOG_PRECOMP = 0x1000, LOG_PROFILE = 0x2000, LOG_FUNC = 0x4000,
	LOG_GPUPROF = 0x8000, TEST_BUILDPROG = 0x10000
};

struct GlobalSettings
{
	int LogLevel, GPU_device, GPUsPerNode;
};

extern GlobalSettings G_Settings;
extern int MyId;
extern char _PID[];

#define EXT_EXDATA "%s.exd.%03i"

#define IFLOGFILE if(G_Settings.LogLevel&(LOG_FILES))
#define IFLOG(x) if(G_Settings.LogLevel&(x))

// prints function name and registers start/end times

#define ENTER  cl_uint _starttime=0; ++_STACKCOUNT; IFLOG(LOG_PROFILE) _starttime=Microseconds(); \
  IFLOG(LOG_FUNC) { printf("%s%*s%s ",_PID, 2*(_STACKCOUNT-1),"< ", __FUNCTION__); fflush(stdout); }\
//-end-macro

#define ENTERp cl_ulong _starttime=Microseconds(); ++_STACKCOUNT; \
IFLOG(LOG_FUNC) { printf("%s%*s%s",_PID,2*(_STACKCOUNT-1),"< ",__FUNCTION__); fflush(stdout); }
//-end-macro

#define RETURN(x) { --_STACKCOUNT; IFLOG(LOG_PROFILE) \
{ _starttime=Microseconds()-_starttime; \
printf("%s%*s%s - %.2fms ",_PID,2*(_STACKCOUNT),"> ",__FUNCTION__,0.001f*(float)_starttime); fflush(stdout);}\
else { IFLOG(LOG_FUNC) { printf("%s%*s%s ", _PID, 2*(_STACKCOUNT),"> ",__FUNCTION__);fflush(stdout);} }\
 return x; }
//-end-macro

#ifdef TARGET_GPU
#define PRINT_CL_ERROR	{	cl_error_occured:		if ( cl_error ) printf("%s%s", _PID, ClErrorName(cl_error)); }
#else
#define PRINT_CL_ERROR	{	cl_error_occured:		if ( cl_error ) printf("%s Error: %i", _PID, cl_error); }
#endif

#ifdef TARGET_WINDOWS
#define SLASH "\\"
#else
#define SLASH "/"
#endif

#define ASSERT(x) if(!(x)) { fprintf(stderr,"\nASSERTION FAILED AT %s %i ", __FILE__,__LINE__);}
