#include <stdio.h>
#include "k_defs.h"
#include "cdata.h"
#include <cstring>
#include <memory>
#include "shaders.h"

//#define regs register
#define regs

ShaderObj* ShaderObj::ShaderBase = NULL;

int ShaderObj::ActiveArch =
#ifdef TARGET_ARM
ShaderObj::FL_ARM;
#else
ShaderObj::FL_INTEL;
#endif

ShaderObj::ShaderObj(const char* nm, int w, int fl, ShaderFn F, int Yg)
	: pF(F), el_wid(w), flags(fl), Y_gran(Yg)
{
	memset(Name, 0, sizeof(Name));
	strncpy(Name, nm, sizeof(Name) - 1);
	pNext = ShaderBase; ShaderBase = this;
}

int ShaderObj::PrintAll() const
{
	int ret = 0;
	ShaderObj* p = ShaderBase;
	while ( p )
	{ p->Print(); p = p->pNext; ++ret; }
	return ret;
}


ShaderObj* GetShader(const char* name, int Flags_Req, int Flags_opt, int width_req)
{
restart:
	Flags_Req |= ShaderObj::ActiveArch;
	ShaderObj* p = ShaderObj::ShaderBase;
	bool found;
	for ( found = false; p && !found; p = p->pNext )
	{
		if ( strncmp(name, p->Name, sizeof(p->Name) - 1) ) continue;
		if ( width_req && (p->el_wid != width_req) ) continue;
		if ( (p->flags & Flags_Req) != Flags_Req ) continue;
		if ( (p->flags & Flags_opt) != Flags_opt ) continue;
		found = true; break;
	}
	if ( found ) { IFLOG(LOG_KRNL) { p->Print(); }	return p; }
	printf("\nShader not found '%s' flags=%X width=%i ", name, Flags_opt, width_req);
	if ( Flags_opt == 0 && width_req == 0 )
	{
		printf("\nAvailable shaders:");
		ShaderObj::ShaderBase->PrintAll();
		return NULL;
	} else { Flags_opt = 0; width_req = 0; goto restart; }
}


#define _NAME_LINE_S(name,s) _UNIQUE_NAME_LINE_S(name, __LINE__,s)
#define _UNIQUE_NAME_LINE_S(a,b,c) _NAME_LINE2(a,b,c)
#define _NAME_LINE2( name, line,suf ) name##line##suf


#define SHADER(name, stride, gran) \
SHADER_ALIGN static void _NAME_LINE_S(name,_f) (cl_ulong pCD, cl_ulong offset, int H, int X0, int Y0, int T); \
  static ShaderObj _NAME_LINE_S(name,_o) (#name,stride,(ShaderObj::FL_2D)|SHADER_DECL_ARCH, _NAME_LINE_S(name,_f), gran); \
SHADER_ALIGN static void _NAME_LINE_S(name,_f) (cl_ulong pCD, cl_ulong offset, int H, int X0, int Y0, int T)
//-end-macro

#define SHADER1(name, stride) \
SHADER_ALIGN static void  _NAME_LINE_S(name,_f) (cl_ulong pCD, cl_ulong offset, int Range, int X0, int _H, int T); \
  static ShaderObj _NAME_LINE_S(name,_o) (#name,stride,(ShaderObj::FL_1D)|SHADER_DECL_ARCH, _NAME_LINE_S(name,_f)); \
SHADER_ALIGN static void  _NAME_LINE_S(name,_f) (cl_ulong pCD, cl_ulong offset, int Range, int X0, int _H, int T)
//-end-macro

#define SHADER_DECL_ARCH (ShaderObj::FL_ARM | ShaderObj::FL_INTEL)

#define CD ((CData*)(pCD))

#define GF_(x) (*(float*)(x))

//#define RANGE_CHECKS

SHADER(SaveExData2D, 1, 1)
{ //printf("\nSaveE X0=%i Y0=%i R=%i T=%i ", X0, Y0, H>>16, T);
	int HH = (H >> 16);
	float*   ptr = (float*)((CD->ExData.HostAddr(Y0, T - 1)));
	cl_uint* bound = (cl_uint*)(CD->ExData.HostAddr(-1, CD->ExData.LinesCommon - 1));
	cl_ulong Data = CD->pU[T & 1];

#ifdef RANGE_CHECKS
	int V = 0, Q = 1;
	if ( !ptr || !bound || !Data || !T ) goto error;
	Q = 2;
	if ( Y0 >= CD->HostSrcs ) goto error;
	Q = 3;
	if ( Y0 >= CD->ExData.SrcsAllocated ) goto error;
	Q = 4;
	if ( (T - 1) >= CD->ExData.LinesPerSrc ) goto error;
	Q = 5;
	if ( ((X0 + HH) * 4) >= CD->ExData.LineWidthB ) goto error;
	Q = 6;
	ptr += X0; bound += X0; Data += CD->DZ*Y0;

	for ( int i = 0; i <= HH; ++i )
	{
		V = bound[i];
		if ( V > CD->FullW*CD->ReqH * 4 ) goto error;
		float* pF = (float*)(Data + bound[i]);
		ptr[i] = *pF;
	}
	return;
error:
	fprintf(stderr, "\nSaveExData OOB! ptr=0x%016llX bound=0x%016llX Data=0x%016llX X0=%i Y0=%i T=%i HH=%i Q=%i V=0x%08X",
					ptr, bound, Data, X0, Y0, T, HH, Q, V);
	fflush(stderr);
	exit(1);
	return;


#else
	ptr += X0; bound += X0; Data += CD->DZ*Y0;
#pragma simd
	for ( int i = 0; i <= HH; ++i )
		ptr[i] = *(float*)(Data + bound[i]);
	return;

error:
	printf("\nSaveExData OOB! ptr=0x%016llX bound=0x%016llX Data=0x%016llX X0=%i Y0=%i T=%i HH=%i",
				 ptr, bound, Data, X0, Y0, T, HH);
	fflush(stdout);
	exit(1);
	return;
#endif	

}

SHADER(SaveBound2D, 1, 1)
{
	//printf("\nSaveB X0=%i Y0=%i R=%i T=%i ", X0, Y0, H>>16, T);
	int HH = H >> 16;
	float*   const ptr = (float*)((CD->BoundCh.HostAddr(Y0, T - 1))) + X0;
	cl_uint* const bound = (cl_uint*)(CD->BoundCh.HostAddr(-1, 0)) + X0;
	cl_ulong const Data = CD->pU[T & 1] + CD->DZ*Y0;
	if ( !ptr || !bound || !Data ) { printf(" B-SKIP! "); return; }
#pragma simd
	for ( int i = 0; i <= HH; ++i )
		ptr[i] = *(float*)(Data + bound[i]);
}

SHADER(LoadBound2D, 1, 1)
{
	//printf("\nLoadB X0=%i Y0=%i R=%i T=%i ", X0, Y0, H>>16, T);
	int HH = H >> 16;
	float*  ptr = (float*)(CD->BoundCh.HostAddr(Y0, T - 1));
	if ( !ptr ) return;
	ptr += X0;
	cl_uint* const bound = (cl_uint*)(CD->BoundCh.HostAddr(-1, 0)) + X0;
	cl_ulong const Data = CD->pU[T & 1] + CD->DZ*Y0;
#pragma simd
	for ( int i = 0; i <= HH; ++i )
		*(float*)(Data + bound[i]) = ptr[i];
}

SHADER(LoadExData2D, 1, 1)
{
//	printf("\nLoadE X0=%i Y0=%i R=%i T=%i ", X0, Y0, H >> 16, T); getchar();
	const int TBase = T - 1;
	if ( (TBase < 1) || (TBase >= CD->NFrames) ) return;

	float*  ptr = ((float*)CD->ExDataFileMaps[Y0 + CD->StartSrc].access);
	if ( !ptr ) return;

	const float rdt2 = 1.0f / (CD->dt*CD->dt);
	const int tsel = T & 1, detlim = CD->NumDetectors - X0 - 1;
	int HH = (H >> 16);
	HH = __min(HH, detlim);

	ptr += X0 + TBase * CD->NumDetectorsA;
	cl_uint* const bound = (cl_uint*)(CD->ExData.HostAddr(-1, CD->ExData.LinesCommon - 1)) + X0;
	float* const NevOut = (float*)(CD->ExData.HostAddr(-1, Y0)) + X0;
	cl_ulong const DataW = CD->pW[tsel] + CD->DZ*Y0;
	cl_ulong const DataU = CD->pU[tsel] + CD->DZ*Y0;
	float nvscale = (CD->NFrames - T)*CD->FadeTime;
	nvscale = CLAMP(nvscale, 0.0f, 1.0f);	

#pragma simd
	for ( int i = 0; i <= HH; ++i )
	{
		float nev = (*(float*)(DataU + bound[i]) - ptr[i])*nvscale;
		*(float*)(DataW + bound[i]) += nev;
		NevOut[i] += nev * nev;
	}
}


SHADER1(BoundaryCond, 1)
{ // index X= (Source<<2) | { up/down/left/right}

	cl_ulong pos0, Wincr;

	if ( _H )
	{ pos0 = CD->pW[T & 1];		Wincr = CD->pW[1 - (T & 1)] - pos0; } else
	{ pos0 = CD->pU[T & 1];		Wincr = CD->pU[1 - (T & 1)] - pos0; }

	for ( int X = X0; X < (X0 + Range); ++X )
	{
		cl_ulong  pos = pos0 + (X >> 2)*CD->DZ;
		int Ydir = (X & 2) ? 4 : CD->DY;

#define NRBC_CORE 				float out = CD->Kuh*GF_(pos) + \
				CD->Kdh*GF_(pos + YY) + 	CD->Kdl*GF_(pDwnLo) + \
				CD->Kmid*((MIDHI = GF_(pos + Wincr)) + (MIDLO = GF_(pos + Wincr + Ydir))) +\
				CD->Kdif * DIFF_;\
				GF_(pos + Wincr + YY) = MIDHI;			GF_(pDwnLo + Wincr) = MIDLO;\
				GF_(pos + Ydir) = out;	   GF_(pos + YY) = out*0.75;\
        GF_(pos+YYY)=0;
				//-end-macro

		if ( X & 2 )
		{ // vertical
			if ( X & 1 )  pos += 4 * (CD->InW - 1);
			else  (Ydir = -Ydir);
			const int YY = Ydir + Ydir, YYY = YY + Ydir, YYYY = YY + YY;
			GF_(pos + Ydir) = GF_(pos + YY) = GF_(pos + YYY) = 0;
			pos += CD->DY;
			for ( int i = CD->InH - 2; i > 0; --i, pos += CD->DY )
			{
				cl_ulong pDwnLo = pos + YYYY;
				float MIDHI, MIDLO,
					DIFF_ = GF_(pDwnLo - CD->DY) + GF_(pDwnLo + CD->DY) +
					GF_(pos - CD->DY) + GF_(pos + CD->DY);
				NRBC_CORE;
			} // for i
			GF_(pos + Ydir) = GF_(pos + YY) = GF_(pos + YYY) = 0;
		} // if vertical
		else
		{// horizontal			
			if ( X & 1 )  pos += CD->DY*(CD->InH - 1);
			else Ydir = -Ydir;
			const int YY = Ydir + Ydir, YYY = YY + Ydir, YYYY = YY + YY;

			GF_(pos + Ydir) = GF_(pos + YY) = GF_(pos + YYY) = 0;
			pos += 4;

#pragma simd
			for ( int i = CD->InW - 2; i > 0; --i, pos += 4 )
			{
				cl_ulong pDwnLo = pos + YYYY;
				float MIDHI, MIDLO,
					DIFF_ = GF_(pDwnLo - 4) + GF_(pDwnLo + 4) +
					GF_(pos - 4) + GF_(pos + 4);
				NRBC_CORE;
			} // for i
			GF_(pos + Ydir) = GF_(pos + YY) = GF_(pos + YYY) = 0;
		} // else horizontal		
	} // for lines
} // fn


#define SET_UADDR(pU0,pU1)	cl_ulong pU0 = CD->pU[(T&1)],		pU1 = CD->pU[1-(T&1)];
#define SET_WADDR(pW0,pW1)	cl_ulong pW0 = CD->pW[(T&1)],		pW1 = CD->pW[1-(T&1)];

#undef SHADER_DECL_ARCH
#define SHADER_DECL_ARCH (ShaderObj::FL_ARM)
#include "shaders_arm.i"

#ifndef TARGET_ARM
#undef SHADER_DECL_ARCH
#define SHADER_DECL_ARCH (ShaderObj::FL_INTEL)
#include "shaders_intel.i"
#endif

#ifdef UNIT_TEST
#include "minicatch.hpp"

TEST_CASE("Shaders_1", "Shaders")
{
	REQUIRE(ShaderObj::ShaderBase->PrintAll() > 5);
}

#endif
