#pragma once
#include "k_defs.h"
#include "shaderobj.h"

#define SHADER_ALIGN

#ifdef __GNUC__
#ifndef TARGET_ARM
#define SHADER_ALIGN   __attribute__ ((force_align_arg_pointer))
#endif
#endif

#ifdef ARM_INTRINSICS

template<int lane> static inline float4 vextq(float4 a, float4 b)
{ float4 r; __asm__("ext %0.16b, %1.16b, %2.16b, #%3" : "=w"(r) : "w"(a), "w"(b), "I"(lane * 4)); return r; }

template<int lane> static inline uint4 vextq(uint4 a, uint4 b)
{ uint4 r; __asm__("ext %0.16b, %1.16b, %2.16b, #%3" : "=w"(r) : "w"(a), "w"(b), "I"(lane * 4)); return r; }

static inline void LDPair(float4& a, float4&b, float* ptr)
{ __asm__("ldp %q0, %q1, %2" : "=w"(a), "=w"(b) : "m"(*ptr)); }

static inline float4 SetLow(float4 a, cl_ulong b)
{ __asm__("ins %0.2d[0], %1" : "+w"(a) : "r"(b)); return a; }

static inline float4 SetHigh(float4 a, cl_ulong b)
{ __asm__("ins %0.2d[1], %1" : "+w"(a) : "r"(b)); return a; }

static inline void  prefetch(cl_ulong p)
{ __asm__ __volatile__("prfm PLDL3KEEP, %0" : : "m"(*(int*)p)); }

// R=A[L]
template<int lane> static inline float4 vdup(float4 a)
{ float4 r; __asm__("dup %0.4s, %1.s[%2]" : "=w"(r) : "w"(a), "I"(lane)); return r; }

// R=A*B[L]
template<int lane> static inline float4 fmul(float4 a, float4 b)
{ float4 r; __asm__("fmul %0.4s, %1.4s, %2.s[%3]" : "=w"(r) : "w"(a), "w"(b), "I"(lane)); return r; }

// A+=B*C
static inline float4 acc_mul(float4 a, float4 b, float4 c)
{ __asm__("fmla %0.4s, %1.4s, %2.4s" : "+w"(a) : "w"(b), "w"(c)); return a; }

// A+=B*C[L]
template<int lane> static inline float4 acc_mul(float4 a, float4 b, float4 c)
{ __asm__("fmla %0.4s, %1.4s, %2.s[%3]" : "+w"(a) : "w"(b), "w"(c), "I"(lane)); return a; }

// A-=B*C
static inline float4 nacc_mul(float4 a, float4 b, float4 c)
{ __asm__("fmls %0.4s, %1.4s, %2.4s" : "+w"(a) : "w"(b), "w"(c)); return a; }

// A-=B*C[L]
template<int lane> static inline float4 nacc_mul(float4 a, float4 b, float4 c)
{ __asm__("fmls %0.4s, %1.4s, %2.s[%3]" : "+w"(a) : "w"(b), "w"(c), "I"(lane)); return a; }

#else // no intrinsics


/*
static inline cl_ulong Rdtsc()
{	cl_uint lo, hi;
	__asm__ __volatile__("rdtsc" : "=a" (lo), "=d" (hi));
	return ((cl_ulong)hi << 32) | lo;
}
*/


static inline float4 vdupq_n_f32(float val)
{
	return  float4( val, val, val, val );
}

#ifdef TARGET_ARM
template<int lane> static inline float4 vextq(float4 a, float4 b)
{
	return float4(
		a[lane], (lane < 3) ? a[lane + 1] : b[lane - 3],
		(lane < 2) ? a[lane + 2] : b[lane - 2],
		(lane < 1) ? a[lane + 3] : b[lane - 1] );
}

#else // on Intel CPU

template<int lane> static inline float4 vextq(float4 a, float4 b)
{ return  reinterpret_f(blend4<lane,lane+1,lane+2,lane+3>(a,b)); }

template<int lane> static inline uint4 vextq(uint4 a, uint4 b)
{ return  (blend4<lane,lane+1,lane+2,lane+3>(a,b)); }

#endif

static inline void LDPair(float4& a, float4&b, float* ptr)
{ a = *(float4*)ptr; b = *(float4*)(ptr + 4); }

static inline float4 SetLow(float4 a, cl_ulong b)
{
	cl_uint tmpa = b & 0xFFFFFFFFUL, tmpb = b >> 32;
	return float4( reinterpret_cast<float&>(tmpa), reinterpret_cast<float&>(tmpb), a[2], a[3] );
}

static inline float4 SetHigh(float4 a, cl_ulong b)
{
	cl_uint tmpa = b & 0xFFFFFFFFUL, tmpb = b >> 32;
	return float4( a[0], a[1], reinterpret_cast<float&>(tmpa), reinterpret_cast<float&>(tmpb) );
}

static inline void  prefetch(cl_ulong p)
{}

// R=A[L]
template<int lane> static inline float4 vdup(float4 a)
{ return float4( a[lane],a[lane],a[lane],a[lane] ); }

// R=A*B[L]
template<int lane> static inline float4 fmul(float4 a, float4 b)
{ return a * b[lane]; }

// A+=B*C
static inline float4& acc_mul(float4& a, float4 b, float4 c)
{ return (a += b * c); }

// A+=B*C[L]
template<int lane> static inline float4& acc_mul(float4& a, float4 b, float4 c)
{ return (a += b * c[lane]); }

// A-=B*C
static inline float4& nacc_mul(float4& a, float4 b, float4 c)
{ return (a -= b * c); }

// A-=B*C[L]
template<int lane> static inline float4& nacc_mul(float4& a, float4 b, float4 c)
{ return (a -= b * c[lane]); }

//vcvtq_n_f32_u32(vshrq_n_u32(AAu, 16), 12);

static inline uint4 vshrq_n_u32(uint4 v, char i) { return (v >> i); }
static inline int4 vshrq_n_s32(int4 v, char i) { return (v >> i); }
static inline uint4 vshlq_n_u32(uint4 v, char i) { return (v << i); }
static inline int4 vshlq_n_s32(int4 v, char i) { return (v << i); }

template<class VV, unsigned char shift> static inline float4 vcvtq_n_f32_x32_(VV v)
{
	constexpr float scale = 1.0f/(float)(1ULL<<shift);
	return float4( v[0],v[1],v[2],v[3] ) *scale;
}

#define vcvtq_n_f32_u32(v,i) vcvtq_n_f32_x32_<uint4,i>(v)
#define vcvtq_n_f32_s32(v,i) vcvtq_n_f32_x32_<int4,i>(v)


#endif // target intel

static inline void LDPair_(float4& a, float4&b, float* ptr)
{ a = *(float4*)ptr; b = *(float4*)(ptr + 4); }

